unit u_map;

interface

uses
  enZGLine;

const
  MAP_GROUP = 1;
  l: TVector3D = (x: 0.577350267;
                  y: 0.577350267;
                  z: 0.577350267);

type
  TVerticle = record
    vec   : TVector3D;
    tx1   : TVector2D;
    tx2   : TVector2D;
  end;

  TPix = record
    R, G, B: byte;
  end;

  TChunk = object
    O_Obj     : TGLOccObj;
    V_Buf     : TGLBuffer;
    bData     : array of array of byte;
    Verticles : array of TVerticle;
    BBox      : TBox;//Bounding box;
    c         : TVector3D;//center;
    Visible   : boolean;

    procedure FillVert(si, sj: integer);
  end;

  TMap = object
    T_Tex: TGLTexture;
    T_Det: TGLTexture;
    I_Buf: TGLBuffer;
    Chunks: array[0..15,0..15] of TChunk;
    Render_I, Render_J: array of integer;
    Chunks_vis: integer;
    iidx: array of integer;
    jidx: array of integer;
    len : array of single;

    procedure Init;
    procedure Render;
    procedure AnalizeChunks;
  end;

var
  bData: array[0..512] of array[0..512] of byte;

var
  map: TMap;
  indexes: array of WORD;
  Filter: byte = TF_ANISOTROPY;  //1
  AntiAlias : boolean = true;    //2
  Cull      : boolean = true;    //3
  Wire      : boolean = false;   //4
  Detail    : boolean = true;    //5
  fog       : boolean = true;    //6

implementation

{ TMap }

procedure TMap.AnalizeChunks;
var
  i, j: integer;
  p: TVector3D;
  l: integer;
  min_: integer;
  fTmp: single;
  iTmp: integer;

  Counter: integer;
begin
  p := cam.GetPos;
  SetLength(iidx,0);
  SetLength(jidx,0);
  SetLength(len,0);
  Counter := 0;

  if not Cull then
  begin
    for i := 0 to 15 do
      for j := 0 to 15 do
        Chunks[i,j].Visible := true;
    exit;
  end;

  for i := 0 to 15 do
    for j := 0 to 15 do
    begin
      Chunks[i,j].Visible := math3D.F_BoxInFrust(Chunks[i,j].BBox);
      if Chunks[i,j].Visible then
      begin
        l := length(iidx);
        SetLength(iidx,l+1);
        SetLength(jidx,l+1);
        SetLength(len,l+1);
        iidx[l] := i;
        jidx[l] := j;
        len[l] := math3D.V_Abs(math3D.V_Sub(p,Chunks[i,j].c));
        inc(Counter);
      end;
    end;

  if Counter <= 1 then exit;

  l := length(len)-1;
  for i := 0 to l do
  begin
    min_ := i;
    for j := i to l do
    begin
      if len[j] < len[min_] then
        min_ := j;
    end;
    if min_ <> i then
    begin
      fTmp := len[min_];
      len[min_] := len[i];
      len[i] := fTmp;

      iTmp := iidx[min_];
      iidx[min_] := iidx[i];
      iidx[i] := iTmp;

      iTmp := jidx[min_];
      jidx[min_] := jidx[i];
      jidx[i] := iTmp;
    end;
  end;

  occ.Enable;
  for i := 0 to l do
    occ.PostQuery(Chunks[iidx[i],jidx[i]].O_Obj);
  occ.Disable;    
  while not occ.IsQueryReady(Chunks[iidx[l],jidx[l]].O_Obj) do begin end;
  for i := 0 to l do
    Chunks[iidx[i],jidx[i]].Visible := occ.GetQueryResult(Chunks[iidx[i],jidx[i]].O_Obj) <> 0;
end;

procedure TMap.Init;
var
  i, j, ci, cj: integer;
  stream: ICustomStream;
  intens: single;
  v: TVector3D;
begin
  stream := fsys.OpenFile('Data\Height.raw');
  stream.Seek(0,FROM_BEGIN);
  stream.Read(bData,513*513);
  stream := nil;

  T_Det := tex.LoadTexture('Data\Det.tga',true,false,MAP_GROUP,'*Det*');

  for i := 0 to 15 do
    for j := 0 to 15 do
    begin
      SetLength(Chunks[i,j].bData,33);
      for ci := 0 to 32 do
      begin
        SetLength(Chunks[i,j].bData[ci],33);
      end;

      for ci := 0 to 32 do
        for cj := 0 to 32 do
          Chunks[i,j].bData[ci,cj] := bData[ci+i*32,cj+j*32];
    end;

  T_Tex := tex.LoadTexture('Data\MapTex.jpg',true,true,MAP_GROUP,'*Map*');

  SetLength(indexes,6144);
  for i := 0 to 31 do
    for j := 0 to 31 do
    begin
      indexes[(j+i*32)*6 + 0] := (j+0) + (i+0)*33;
      indexes[(j+i*32)*6 + 1] := (j+1) + (i+0)*33;
      indexes[(j+i*32)*6 + 2] := (j+1) + (i+1)*33;

      indexes[(j+i*32)*6 + 3] := (j+0) + (i+0)*33;
      indexes[(j+i*32)*6 + 4] := (j+1) + (i+1)*33;
      indexes[(j+i*32)*6 + 5] := (j+0) + (i+1)*33;
    end;

  I_Buf := vbo.NewID(6144*2,BUF_ELEMENT,addr(indexes[0]),'I_Buf');
  for i := 0 to 15 do
    for j := 0 to 15 do
      Chunks[i,j].FillVert(i*32,j*32);
 end;

procedure TMap.Render;
var
  i, j: integer;
begin
  math3D.BuildFrustum;
  AnalizeChunks;
  ogl.CullFacing(FACE_BACK);
  tex.EnableTextures2D;
  if not wire then
  begin
    tex.AssignToTexUnit(T_Tex,0,ENV_REPLACE);
    if Detail then
    begin
      tex.AssignToTexUnit(T_Det,1,ENV_BLEND);
    end;
  end;

  vbo.EnableArray(ARR_VERTEX);
  vbo.EnableArray(ARR_TEXCOORD);
  vbo.EnableArray(ARR_TEXCOORD);
  Chunks_vis := 0;

  vbo.IEnable(I_Buf);
  for i := 0 to 15 do
    for j := 0 to 15 do
    begin
      if not Chunks[i,j].Visible then continue;
      Inc(Chunks_vis);
      vbo.VEnable(Chunks[i,j].V_Buf);
      vbo.VertexPointer(28,0,3);
      vbo.TexCoordPointer(28,12,2,0);
      vbo.TexCoordPointer(28,20,2,1);
      vbo.RenderBuffer(TRIANGLES,6144,0);
      vbo.VDisable;
    end;

  vbo.IDisable;
  vbo.DisableArray(ARR_TEXCOORD);
  vbo.DisableArray(ARR_TEXCOORD);
  vbo.DisableArray(ARR_VERTEX);
  if not wire then
  begin
    tex.Disable(0);
    if Detail then
      tex.Disable(1);
  end;
  tex.DisableTextures2D;
  ogl.CullFacing(FACE_NONE);  
end;

{ TChunk }

procedure TChunk.FillVert(si, sj: integer);
var
  i, j: integer;
begin
  SetLength(Verticles,1089);
  BBox.min := math3D.NewVector3D( 100000, 100000, 100000);
  BBox.max := math3D.NewVector3D(-100000,-100000,-100000);
  for i := 0 to 32 do
    for j := 0 to 32 do
    begin
      Verticles[j+i*33].vec.x := 5*(si + i);
      Verticles[j+i*33].vec.z := 5*(sj + j);
      Verticles[j+i*33].vec.y := 150*bData[i,j]/255;

      if BBox.min.x > Verticles[j+i*33].vec.x then BBox.min.x := Verticles[j+i*33].vec.x;
      if BBox.max.x < Verticles[j+i*33].vec.x then BBox.max.x := Verticles[j+i*33].vec.x;

      if BBox.min.y > Verticles[j+i*33].vec.y then BBox.min.y := Verticles[j+i*33].vec.y;
      if BBox.max.y < Verticles[j+i*33].vec.y then BBox.max.y := Verticles[j+i*33].vec.y;

      if BBox.min.z > Verticles[j+i*33].vec.z then BBox.min.z := Verticles[j+i*33].vec.z;
      if BBox.max.z < Verticles[j+i*33].vec.z then BBox.max.z := Verticles[j+i*33].vec.z;

      Verticles[j+i*33].tx1.x := (si + i)/513;
      Verticles[j+i*33].tx1.y := (sj + j)/513;

      Verticles[j+i*33].tx2 := math3D.NewVector2D(i/8,j/8);
    end;

  c := math3D.V_Mul(math3D.V_Add(BBox.min,BBox.max),0.5);

  V_Buf := vbo.NewID(1089*sizeof(TVerticle),BUF_DATA,addr(Verticles[0]),'V_Buf');
  SetLength(Verticles,0);
  O_Obj := occ.NewID(BBox);
end;

end.
